from __future__ import annotations
from typing import Set, List, Dict, Tuple, Optional, TypeVar
from enum import Enum
import enum
from copy import deepcopy
from collections import deque


class Composition(Enum):
    Union = enum.auto()
    Intersection = enum.auto()
    Subtraction = enum.auto()


class Terminal:
    def __init__(self, name: str):
        self.name = name

    def __eq__(self, obj):
        if isinstance(obj, Terminal):
            return obj.name == self.name
        return False

    def __hash__(self):
        return hash(self.name)


class Eps:
    def __init__(self):
        pass


class Nonterminal:
    def __init__(self, name: str):
        self.name = name

    def __eq__(self, obj):
        if isinstance(obj, Nonterminal):
            return obj.name == self.name
        return False

    def __hash__(self):
        return hash(self.name)


class State:
    def __init__(self, name: str):
        self.name = name

    def __eq__(self, obj):
        if isinstance(obj, State):
            return obj.name == self.name
        return False

    def __hash__(self):
        return hash(self.name)


class REG:
    Rules = Dict[Nonterminal, Set[Tuple[Terminal, Optional[Nonterminal]]]]
    type_var = TypeVar('type_var')

    def set_or_none(self, out : Optional[Set[type_var]]) -> Set[type_var]:
        return set() if out is None else out

    def dict_or_none(self, out : Optional[Dict[type_var]]) -> Set[type_var]:
        return dict() if out is None else out

    def __init__(self, nonterminals: Set[Nonterminal] = None,
                 terminals: Set[Terminal] = None,
                 rules: Rules = None,
                 init: Nonterminal = None):
        self.nonterminals = self.set_or_none(nonterminals)
        self.terminals = self.set_or_none(terminals)
        self.rules = self.dict_or_none(rules)
        self.init = init

    def reg_to_nfa(self) -> NFA:
        dfa = DFA(set(), set(), {}, None, set())
        for nonterminal in self.nonterminals:
            state = State(nonterminal.name)
            dfa.states.add(state)

        dfa.terminals = self.terminals
        dfa.init = State(self.init.name)
        final = State("new_final")  # TODO assert that name is unique, as for hell in make_total
        dfa.states.add(final)
        dfa.final.add(final)

        for nonterminal in self.rules:
            for rule in self.rules[nonterminal]:

                # rule A -> a becomes d(A, a) = final
                if type(rule) == Terminal: # TODO and a is not \eps
                    if (nonterminal, rule) in dfa.transition:
                        dfa.transition[nonterminal, rule].add(final)
                    else:
                        dfa.transition[nonterminal, rule] = {final}

                # rule A -> aB becomes d(A, a) = B
                elif (nonterminal, rule[0]) in dfa.transition:
                    dfa.transition[nonterminal, rule[0]].add(rule[1])
                else:
                    dfa.transition[nonterminal, rule[0]] = {rule[1]}

        # TODO if init -> \eps: dfa.final.add(dfa.init)
        # I need to know how to treat \eps
        return dfa


class DFA:
    Transition = Dict[Tuple[State, Terminal], State]
    type_var = TypeVar('type_var')

    def set_or_none(self, out : Optional[Set[type_var]]) -> Set[type_var]:
        return set() if out is None else out

    def dict_or_none(self, out : Optional[Dict[type_var]]) -> Set[type_var]:
        return dict() if out is None else out

    def __init__(self, states: Set[State] = None,
                 terminals: Set[Terminal] = None,
                 transition: Transition = None,
                 init: State = None,
                 final: Set[State] = None):
        # TODO learn how to assert
        # assert all or nothing
        # assert init exists
        # assert final is subset of states
        # valid automaton has min one state!
        self.states = self.set_or_none(states)
        self.terminals = self.set_or_none(terminals)
        self.transition = self.dict_or_none(transition)
        self.init = init
        self.final = self.set_or_none(final)

    def total(self) -> DFA:
        if self.is_total():
            return self

        dfa = deepcopy(self)
        # total: Transition
        total = {}

        # check: consider import itertools
        hell_id = 0
        hell = State(str(hell_id))
        while hell in dfa.states:
            hell_id += 1
            hell.name = str(hell_id)
        dfa.states.add(hell)

        for state in dfa.states:
            for terminal in dfa.terminals:
                if (state, terminal) in dfa.transition:
                    total[state, terminal] = dfa.transition[state, terminal]
                else:
                    total[state, terminal] = hell

        dfa.transition = total
        return dfa

    def is_total(self) -> bool:
        for state in self.states:
            for terminal in self.terminals:
                if (state, terminal) not in self.transition:
                    return False
        return True

    def accepts(self, word: str) -> bool:
        state = self.init
        for character in word:
            terminal = Terminal(character)
            if (state, terminal) not in self.transition:
                return False
            state = self.transition[state, terminal]
        return state in self.final

    def complement(self) -> DFA:
        dfa = deepcopy(self)
        if not dfa.is_total():
            dfa = dfa.total()

        dfa.final = dfa.states.difference(dfa.final)
        return dfa

    @staticmethod
    def composition(dfa_1: DFA, dfa_2: DFA, operation: Composition) -> DFA:

        # 2019 IB102 slides want both total
        if not dfa_1.is_total():
            dfa_1 = dfa_1.total()
        if not dfa_2.is_total():
            dfa_2 = dfa_2.total()

        dfa = DFA()

        # new terminals
        dfa.terminals = dfa_1.terminals.union(dfa_2.terminals)
        # ATTENTION: terminals may differ -> make total at the end?

        # new init
        dfa.init = State(f"{dfa_1.init.name}_{dfa_2.init.name}")
        dfa.states.add(dfa.init)

        # new final TODO also outside
        for state_1 in dfa_1.final:
            for state_2 in dfa_2.final:
                state = State(f"{state_1.name}_{state_2.name}")
                dfa.states.add(state)

                if operation == Composition.Union:
                    dfa.final.add(state)

                elif operation == Composition.Intersection and \
                        state_1 in dfa_1.final and \
                        state_2 in dfa_2.final:
                    dfa.final.add(state)

                elif operation == Composition.Subtraction and \
                        state_2 not in dfa_2.final:
                    dfa.final.add(state)

        # rest of states
        for state_1 in dfa_1.states:
            for state_2 in dfa_2.states:
                state = State(f"{state_1.name}_{state_2.name}")  # TODO avoid conflicts!
                dfa.states.add(state)

                # new transition
                for terminal in dfa.terminals:
                    if (state_1, terminal) in dfa_1.transition and \
                            (state_2, terminal) in dfa_2.transition:
                        move_1 = dfa_1.transition[state_1, terminal]
                        move_2 = dfa_2.transition[state_2, terminal]
                        move = State(move_1.name + "_" + move_2.name)
                        dfa.transition[state, terminal] = move

        return dfa

    @staticmethod
    def union(dfa_1: DFA, dfa_2: DFA) -> DFA:
        return DFA.composition(dfa_1, dfa_2, Composition.Union)

    @staticmethod
    def intersection(dfa_1: DFA, dfa_2: DFA) -> DFA:
        return DFA.composition(dfa_1, dfa_2, Composition.Intersection)

    @staticmethod
    def subtraction(dfa_1: DFA, dfa_2: DFA) -> DFA:
        return DFA.composition(dfa_1, dfa_2, Composition.Subtraction)

    def eliminate_unreachable(self) -> DFA:
        dfa = deepcopy(self)

        previous = set()
        actual = {self.init}

        while previous != actual:
            previous = deepcopy(actual)

            for state in previous:
                for terminal in dfa.terminals:
                    if (state, terminal) in self.transition:
                        actual.add(self.transition[state, terminal])

        dfa.states = actual
        dfa.final = dfa.final.intersection(actual)
        for (state, terminal) in self.transition:
            if state not in actual:
                del dfa.transition[state, terminal]

        return dfa


    # TODO sorted(set, key = lambda x: x.name) sorted_terminals
    def sort_terminals(self) -> List[Terminal]:
        to_sort = deepcopy(self.terminals)
        to_sort = list(to_sort)
        names = []
        for terminal in to_sort:
            names.append((terminal.name, terminal))
        names.sort()
        sorted_terminals = []
        for name, terminal in names:
            sorted_terminals.append(terminal)
        return sorted_terminals

    def minimize(self) -> DFA:
        dfa = deepcopy(self)
        dfa = dfa.eliminate_unreachable()
        dfa = dfa.total()

        terminals = self.sort_terminals()  # bc I want to iterate always in same order

        previous : Optional[Dict[State, int]] = None # TODO check this also at other places
        actual = dict()
        for state in dfa.states:
            if state in dfa.final:
                actual[state] = 1
            else:
                actual[state] = 0

        while previous != actual:
            previous = deepcopy(actual)
            patterns = dict()
            actual = dict()

            for state in dfa.states:
                pattern = [previous[state]]
                for terminal in terminals:
                    result = dfa.transition[state, terminal]
                    pattern.append(previous[result])
                pattern = tuple(pattern)

                if pattern in patterns:
                    patterns[pattern].add(state)
                else:
                    patterns[pattern] = {state}

            i = 0
            for states in patterns.values():
                for state in states:
                    actual[state] = i
                i += 1

        new_transition = {}
        dfa.states = set()
        dfa.init = None
        dfa.final = set()
        for state in actual:
            new_state = State("new_" + str(actual[state]))
            dfa.states.add(new_state)

            for terminal in terminals:
                if not (new_state, terminal) in new_transition:
                    originally = dfa.transition[state, terminal]
                    now = State("new_" + str(actual[originally]))
                    new_transition[new_state, terminal] = now

            if state == self.init:
                dfa.init = new_state
            if state in self.final:
                dfa.final.add(new_state)

        dfa.transition = new_transition
        return dfa

    def rename_state(self, rename: State, name: str):
        new_state = State(name)
        self.states.add(new_state)
        self.states.remove(rename)
        transition = deepcopy(self.transition)

        if rename == self.init:
            self.init = new_state
        if rename in self.final:
            self.final.add(new_state)
            self.final.remove(rename)

        for terminal in self.terminals:
            if (rename, terminal) in self.transition:
                transition[new_state, terminal] = self.transition[rename, terminal]
                del transition[rename, terminal]

        for state, terminal in transition:
            if transition[state, terminal] == rename:
                del transition[state, terminal]
                transition[state, terminal] = new_state

        self.transition = transition

    def canonize(self) -> DFA:
        # TODO total?

        dfa = deepcopy(self)
        remain = deepcopy(self.states)
        terminals = dfa.sort_terminals()

        i = 0
        queue = [dfa.init] # TODO deque
        while len(queue) > 0:
            actual = queue.pop(0)

            for terminal in terminals:
                state = dfa.transition[actual, terminal]
                if state in remain:
                    queue.append(state)
                    remain.remove(state)

            dfa.rename_state(actual, str(i))
            i += 1
        return dfa

    def dfa_to_reg(self) -> REG:
        reg = REG(set(), set(), {}, None)

        for state in self.states:
            nonterminal = Nonterminal(state.name)
            reg.nonterminals.add(nonterminal)

        reg.terminals = self.terminals
        init = Nonterminal("new_init")  # TODO assert that name is unique, as for hell in make_total
        reg.init = init
        reg.nonterminals.add(init)

        for state, terminal in self.transition:
            move = self.transition[state, terminal]
            if state == self.init:
                nonterminal2 = Nonterminal(move.name)
                if init in reg.rules:
                    reg.rules[init].add((terminal, nonterminal2))
                else:
                    reg.rules[init] = {(terminal, nonterminal2)}
                if move in self.final:
                    reg.rules[init].add((terminal))
            else:
                nonterminal1 = Nonterminal(state.name)
                nonterminal2 = Nonterminal(move.name)
                if nonterminal1 in reg.rules:
                    reg.rules[nonterminal1].add((terminal, nonterminal2))
                else:
                    reg.rules[nonterminal1] = {(terminal, nonterminal2)}
                if move in self.final:
                    reg.rules[nonterminal1].add((terminal))

        if self.init in self.final:
            pass # TODO eg.rules[init].add((\eps))

        return reg

    def is_empty(self) -> bool:
        reached = deque(self.init)
        reachable = set(self.init)
        while len(reached) > 0:
            actual = reachable.popleft()
            for terminal in self.terminals:
                if self.transition[actual, terminal] is not None and self.transition[actual, terminal] not in reachable:
                    reached.append(self.transition[actual, terminal])
                    reachable.add(self.transition[actual, terminal])

        return len(reachable.intersection(self.final)) > 0

    # Ha!
    def is_universal(self):
        return self.is_empty(self.complement())


class NFA:
    def __init__(self, states: Set[State],
                 terminals: Set[Terminal],
                 transition: Dict[Tuple[State, Terminal], Set[State]],
                 init: State,
                 final: Set[State]):
        self.states = states
        self.terminals = terminals
        self.transition = transition
        self.init = init
        self.final = final

    def determinize(self) -> DFA:
        states = set(frozenset(self.init))
        transition = {}
        final = set()
        done = set()

        todo = states.difference(done)
        while len(todo) > 0:
            subset = iter(todo).next()# arbitrary element from set
            if subset.difference(self.final) > 0:
                final.add(subset)
            for terminal in self.terminals:
                new_subset = set()
                for state in subset:
                    if (state, terminal) in self.transition:
                        new_subset.add(self.transition[state, terminal])
                states.add(new_subset)
                transition[subset, terminal] = new_subset

            done.add(subset)

        dfa = DFA(set(), self.terminals, {}, self.init, set())
        for state in states:
            dfa.states.add(self.unset(state))
        for state in final:
            dfa.final.add(self.unset(state))
        for state, terminal in transition:
            dfa.transition[self.unset(state), terminal] = self.unset(self.transition[state, terminal])

        # states from sets


    def unset(self, states : Set[State]) -> State:
        return State('_'.join(states.name))
